@@ -1,11 +1,11 @@
 from __future__ import annotations
 
 from collections import deque
-from typing import Deque, Dict, List, Optional, Set, Tuple
+from functools import cached_property
+from typing import Any, Deque, Dict, FrozenSet, Iterable, List, Optional, Set, Tuple
 
 import numpy as np
 
-from sc2.cache import property_immutable_cache, property_mutable_cache
 from sc2.pixel_map import PixelMap
 from sc2.player import Player, Race
 from sc2.position import Point2, Rect, Size
@@ -13,40 +13,40 @@ from sc2.position import Point2, Rect, Size
 
 class Ramp:
 
-    def __init__(self, points: Set[Point2], game_info: GameInfo):
+    def __init__(self, points: FrozenSet[Point2], game_info: GameInfo):
         """
         :param points:
         :param game_info:
         """
-        self._points: Set[Point2] = points
+        self.cache: Dict[str, Any] = {}
+        self._points: FrozenSet[Point2] = points
         self.__game_info = game_info
         # Tested by printing actual building locations vs calculated depot positions
         self.x_offset = 0.5
         self.y_offset = 0.5
-        # Can this be removed?
         self.cache = {}
 
-    @property_immutable_cache
+    @cached_property
     def _height_map(self):
         return self.__game_info.terrain_height
 
-    @property_immutable_cache
+    @cached_property
     def _placement_grid(self):
         return self.__game_info.placement_grid
 
-    @property_immutable_cache
+    @cached_property
     def size(self) -> int:
         return len(self._points)
 
     def height_at(self, p: Point2) -> int:
         return self._height_map[p]
 
-    @property_mutable_cache
-    def points(self) -> Set[Point2]:
+    @cached_property
+    def points(self) -> FrozenSet[Point2]:
         return self._points.copy()
 
-    @property_mutable_cache
-    def upper(self) -> Set[Point2]:
+    @cached_property
+    def upper(self) -> FrozenSet[Point2]:
         """ Returns the upper points of a ramp. """
         current_max = -10000
         result = set()
@@ -57,27 +57,27 @@ class Ramp:
                 result = {p}
             elif height == current_max:
                 result.add(p)
-        return result
+        return frozenset(result)
 
-    @property_mutable_cache
-    def upper2_for_ramp_wall(self) -> Set[Point2]:
+    @cached_property
+    def upper2_for_ramp_wall(self) -> FrozenSet[Point2]:
         """ Returns the 2 upper ramp points of the main base ramp required for the supply depot and barracks placement properties used in this file. """
         if len(self.upper) > 5:
             # NOTE: this was way too slow on large ramps
-            return set()  # HACK: makes this work for now
+            return frozenset()  # HACK: makes this work for now
             # FIXME: please do
 
-        return set(sorted(list(self.upper), key=lambda x: x.distance_to_point2(self.bottom_center), reverse=True)[:2])
+        return frozenset(sorted(self.upper, key=lambda x: x.distance_to_point2(self.bottom_center), reverse=True)[:2])
 
-    @property_immutable_cache
+    @cached_property
     def top_center(self) -> Point2:
         upper = self.upper
         length = len(upper)
         pos = Point2((sum(p.x for p in upper) / length, sum(p.y for p in upper) / length))
         return pos
 
-    @property_mutable_cache
-    def lower(self) -> Set[Point2]:
+    @cached_property
+    def lower(self) -> FrozenSet[Point2]:
         current_min = 10000
         result = set()
         for p in self._points:
@@ -87,37 +87,37 @@ class Ramp:
                 result = {p}
             elif height == current_min:
                 result.add(p)
-        return result
+        return frozenset(result)
 
-    @property_immutable_cache
+    @cached_property
     def bottom_center(self) -> Point2:
         lower = self.lower
         length = len(lower)
         pos = Point2((sum(p.x for p in lower) / length, sum(p.y for p in lower) / length))
         return pos
 
-    @property_immutable_cache
+    @cached_property
     def barracks_in_middle(self) -> Optional[Point2]:
         """ Barracks position in the middle of the 2 depots """
         if len(self.upper) not in {2, 5}:
             return None
         if len(self.upper2_for_ramp_wall) == 2:
-            points = self.upper2_for_ramp_wall
+            points = set(self.upper2_for_ramp_wall)
             p1 = points.pop().offset((self.x_offset, self.y_offset))
             p2 = points.pop().offset((self.x_offset, self.y_offset))
             # Offset from top point to barracks center is (2, 1)
             intersects = p1.circle_intersection(p2, 5**0.5)
-            anyLowerPoint = next(iter(self.lower))
-            return max(intersects, key=lambda p: p.distance_to_point2(anyLowerPoint))
+            any_lower_point = next(iter(self.lower))
+            return max(intersects, key=lambda p: p.distance_to_point2(any_lower_point))
         raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
 
-    @property_immutable_cache
+    @cached_property
     def depot_in_middle(self) -> Optional[Point2]:
         """ Depot in the middle of the 3 depots """
         if len(self.upper) not in {2, 5}:
             return None
         if len(self.upper2_for_ramp_wall) == 2:
-            points = self.upper2_for_ramp_wall
+            points = set(self.upper2_for_ramp_wall)
             p1 = points.pop().offset((self.x_offset, self.y_offset))
             p2 = points.pop().offset((self.x_offset, self.y_offset))
             # Offset from top point to depot center is (1.5, 0.5)
@@ -130,25 +130,25 @@ class Ramp:
             return max(intersects, key=lambda p: p.distance_to_point2(anyLowerPoint))
         raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
 
-    @property_mutable_cache
-    def corner_depots(self) -> Set[Point2]:
+    @cached_property
+    def corner_depots(self) -> FrozenSet[Point2]:
         """ Finds the 2 depot positions on the outside """
         if not self.upper2_for_ramp_wall:
-            return set()
+            return frozenset()
         if len(self.upper2_for_ramp_wall) == 2:
-            points = self.upper2_for_ramp_wall
+            points = set(self.upper2_for_ramp_wall)
             p1 = points.pop().offset((self.x_offset, self.y_offset))
             p2 = points.pop().offset((self.x_offset, self.y_offset))
             center = p1.towards(p2, p1.distance_to_point2(p2) / 2)
             depotPosition = self.depot_in_middle
             if depotPosition is None:
-                return set()
+                return frozenset()
             # Offset from middle depot to corner depots is (2, 1)
             intersects = center.circle_intersection(depotPosition, 5**0.5)
             return intersects
         raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
 
-    @property_immutable_cache
+    @cached_property
     def barracks_can_fit_addon(self) -> bool:
         """ Test if a barracks can fit an addon at natural ramp """
         # https://i.imgur.com/4b2cXHZ.png
@@ -156,7 +156,7 @@ class Ramp:
             return self.barracks_in_middle.x + 1 > max(self.corner_depots, key=lambda depot: depot.x).x
         raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
 
-    @property_immutable_cache
+    @cached_property
     def barracks_correct_placement(self) -> Optional[Point2]:
         """ Corrected placement so that an addon can fit """
         if self.barracks_in_middle is None:
@@ -164,11 +164,10 @@ class Ramp:
         if len(self.upper2_for_ramp_wall) == 2:
             if self.barracks_can_fit_addon:
                 return self.barracks_in_middle
-            else:
-                return self.barracks_in_middle.offset((-2, 0))
+            return self.barracks_in_middle.offset((-2, 0))
         raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
 
-    @property_immutable_cache
+    @cached_property
     def protoss_wall_pylon(self) -> Optional[Point2]:
         """
         Pylon position that powers the two wall buildings and the warpin position.
@@ -182,7 +181,7 @@ class Ramp:
         direction = self.barracks_in_middle.negative_offset(middle)
         return middle + 6 * direction
 
-    @property_mutable_cache
+    @cached_property
     def protoss_wall_buildings(self) -> List[Point2]:
         """
         List of two positions for 3x3 buildings that form a wall with a spot for a one unit block.
@@ -203,7 +202,7 @@ class Ramp:
             return [wall1, wall2]
         raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
 
-    @property_immutable_cache
+    @cached_property
     def protoss_wall_warpin(self) -> Optional[Point2]:
         """
         Position for a unit to block the wall created by protoss_wall_buildings.
@@ -239,7 +238,7 @@ class GameInfo:
         self.playable_area = Rect.from_proto(self._proto.start_raw.playable_area)
         self.map_center = self.playable_area.center
         self.map_ramps: List[Ramp] = None  # Filled later by BotAI._prepare_first_step
-        self.vision_blockers: Set[Point2] = None  # Filled later by BotAI._prepare_first_step
+        self.vision_blockers: FrozenSet[Point2] = None  # Filled later by BotAI._prepare_first_step
         self.player_races: Dict[int, Race] = {
             p.player_id: p.race_actual or p.race_requested
             for p in self._proto.player_info
@@ -247,7 +246,7 @@ class GameInfo:
         self.start_locations: List[Point2] = [Point2.from_proto(sl) for sl in self._proto.start_raw.start_locations]
         self.player_start_location: Point2 = None  # Filled later by BotAI._prepare_first_step
 
-    def _find_ramps_and_vision_blockers(self) -> Tuple[List[Ramp], Set[Point2]]:
+    def _find_ramps_and_vision_blockers(self) -> Tuple[List[Ramp], FrozenSet[Point2]]:
         """Calculate points that are pathable but not placeable.
         Then divide them into ramp points if not all points around the points are equal height
         and into vision blockers if they are."""
@@ -265,12 +264,12 @@ class GameInfo:
             map_area.height and self.placement_grid[(a, b)] == 0
         ]
         # divide points into ramp points and vision blockers
-        rampPoints = [point for point in points if not equal_height_around(point)]
-        visionBlockers = set(point for point in points if equal_height_around(point))
-        ramps = [Ramp(group, self) for group in self._find_groups(rampPoints)]
-        return ramps, visionBlockers
+        ramp_points = [point for point in points if not equal_height_around(point)]
+        vision_blockers = frozenset(point for point in points if equal_height_around(point))
+        ramps = [Ramp(group, self) for group in self._find_groups(ramp_points)]
+        return ramps, vision_blockers
 
-    def _find_groups(self, points: Set[Point2], minimum_points_per_group: int = 8):
+    def _find_groups(self, points: FrozenSet[Point2], minimum_points_per_group: int = 8) -> Iterable[FrozenSet[Point2]]:
         """
         From a set of points, this function will try to group points together by
         painting clusters of points in a rectangular map using flood fill algorithm.
@@ -280,26 +279,26 @@ class GameInfo:
         NOT_COLORED_YET = -1
         map_width = self.pathing_grid.width
         map_height = self.pathing_grid.height
-        currentColor: int = NOT_COLORED_YET
+        current_color: int = NOT_COLORED_YET
         picture: List[List[int]] = [[-2 for _ in range(map_width)] for _ in range(map_height)]
 
         def paint(pt: Point2) -> None:
-            picture[pt.y][pt.x] = currentColor
+            picture[pt.y][pt.x] = current_color
 
         nearby = [(a, b) for a in [-1, 0, 1] for b in [-1, 0, 1] if a != 0 or b != 0]
 
         remaining: Set[Point2] = set(points)
         for point in remaining:
             paint(point)
-        currentColor = 1
+        current_color = 1
         queue: Deque[Point2] = deque()
         while remaining:
-            currentGroup: Set[Point2] = set()
+            current_group: Set[Point2] = set()
             if not queue:
                 start = remaining.pop()
                 paint(start)
                 queue.append(start)
-                currentGroup.add(start)
+                current_group.add(start)
             while queue:
                 base: Point2 = queue.popleft()
                 for offset in nearby:
@@ -312,6 +311,6 @@ class GameInfo:
                     remaining.discard(point)
                     paint(point)
                     queue.append(point)
-                    currentGroup.add(point)
-            if len(currentGroup) >= minimum_points_per_group:
-                yield currentGroup
+                    current_group.add(point)
+            if len(current_group) >= minimum_points_per_group:
+                yield frozenset(current_group)
