@@ -19,6 +19,7 @@ def _sign(num):
 
 
 class Pointlike(tuple):
+
     @property
     def position(self) -> Pointlike:
         return self
@@ -72,13 +73,14 @@ class Pointlike(tuple):
         """This function assumes the 2d distance is meant
 
         :param ps:"""
-        assert ps, f"ps is empty"
+        assert ps, "ps is empty"
+        # pylint: disable=W0108
         return min(ps, key=lambda p: self.distance_to(p))
 
     def distance_to_closest(self, ps: Union[Units, Iterable[Point2]]) -> float:
         """This function assumes the 2d distance is meant
         :param ps:"""
-        assert ps, f"ps is empty"
+        assert ps, "ps is empty"
         closest_distance = math.inf
         for p2 in ps:
             p2 = p2.position
@@ -91,14 +93,15 @@ class Pointlike(tuple):
         """This function assumes the 2d distance is meant
 
         :param ps: Units object, or iterable of Unit or Point2"""
-        assert ps, f"ps is empty"
+        assert ps, "ps is empty"
+        # pylint: disable=W0108
         return max(ps, key=lambda p: self.distance_to(p))
 
     def distance_to_furthest(self, ps: Union[Units, Iterable[Point2]]) -> float:
         """This function assumes the 2d distance is meant
 
         :param ps:"""
-        assert ps, f"ps is empty"
+        assert ps, "ps is empty"
         furthest_distance = -math.inf
         for p2 in ps:
             p2 = p2.position
@@ -144,14 +147,16 @@ class Pointlike(tuple):
     def __eq__(self, other):
         try:
             return all(abs(a - b) <= EPSILON for a, b in itertools.zip_longest(self, other, fillvalue=0))
-        except:
+        except TypeError:
             return False
 
     def __hash__(self):
         return hash(tuple(self))
 
 
+# pylint: disable=R0904
 class Point2(Pointlike):
+
     @classmethod
     def from_proto(cls, data) -> Point2:
         """
@@ -201,14 +206,14 @@ class Point2(Pointlike):
     def to3(self) -> Point3:
         return Point3((*self, 0))
 
-    def offset(self, off):
-        return Point2((self[0] + off[0], self[1] + off[1]))
+    def offset(self, p: Point2):
+        return Point2((self[0] + p[0], self[1] + p[1]))
 
     def random_on_distance(self, distance):
         if isinstance(distance, (tuple, list)):  # interval
             distance = distance[0] + random.random() * (distance[1] - distance[0])
 
-        assert distance > 0, f"Distance is not greater than 0"
+        assert distance > 0, "Distance is not greater than 0"
         angle = random.random() * 2 * math.pi
 
         dx, dy = math.cos(angle), math.sin(angle)
@@ -231,7 +236,7 @@ class Point2(Pointlike):
 
         :param p:
         :param r:"""
-        assert self != p, f"self is equal to p"
+        assert self != p, "self is equal to p"
         distanceBetweenPoints = self.distance_to(p)
         assert r >= distanceBetweenPoints / 2
         # remaining distance from center towards the intersection, using pythagoras
@@ -329,8 +334,9 @@ class Point2(Pointlike):
 
 
 class Point3(Point2):
+
     @classmethod
-    def from_proto(cls, data):
+    def from_proto(cls, data) -> Point3:
         """
         :param data:
         """
@@ -359,6 +365,7 @@ class Point3(Point2):
 
 
 class Size(Point2):
+
     @property
     def width(self) -> float:
         return self[0]
@@ -369,6 +376,7 @@ class Size(Point2):
 
 
 class Rect(tuple):
+
     @classmethod
     def from_proto(cls, data):
         """
